import os
from collections import defaultdict

from fabric.api import env
import logging

from cassandra_backups.snapshotting import (SnapshotCollection,
                                            BackupWorker,
                                            Snapshot,
                                            RestoreWorker)
from utils import get_s3_connection_host, add_s3_arguments, base_parser

env.use_ssh_config = True


def _set_user_env_vars(args):
    if args.user:
        env.user = args.user
    if args.sudo_user:
        env.sudo_user = args.sudo_user
    if args.password:
        env.password = args.password
    if args.sshkey:
        env.key_filename = args.sshkey
    if args.sshport:
        env.port = args.sshport


def run_backup(args):

    _set_user_env_vars(args)
    env.hosts = args.hosts.split(',')
    env.keyspaces = args.keyspaces.split(',') if args.keyspaces else None

    if args.new_snapshot:
        create_snapshot = True
    else:
        existing_snapshot = SnapshotCollection(
            args.aws_access_key_id,
            args.aws_secret_access_key,
            args.s3_base_path,
            args.s3_bucket_name,
            get_s3_connection_host(args.s3_bucket_region)
        ).get_snapshot_for(
            hosts=env.hosts,
            keyspaces=env.keyspaces,
            table=args.table,
            name=Snapshot.make_snapshot_name()
        )
        create_snapshot = existing_snapshot is None

    worker = BackupWorker(
        aws_access_key_id=args.aws_access_key_id,
        aws_secret_access_key=args.aws_secret_access_key,
        s3_bucket_region=args.s3_bucket_region,
        s3_ssenc=args.s3_ssenc,
        s3_connection_host=get_s3_connection_host(args.s3_bucket_region),
        cassandra_conf_path=args.cassandra_conf_path,
        cassandra_tools_bin_dir=args.cassandra_tools_bin_dir,
        cqlsh_user=args.cqlsh_user,
        cqlsh_password=args.cqlsh_password,
        backup_schema=args.backup_schema,
        buffer_size=args.buffer_size,
        use_sudo=args.use_sudo,
        connection_pool_size=args.connection_pool_size,
        exclude_tables=args.exclude_tables,
        reduced_redundancy=args.reduced_redundancy,
        rate_limit=args.rate_limit,
        quiet=args.quiet
    )
    if create_snapshot:
        logging.info("Make a new snapshot")
        snapshot = Snapshot(
            base_path=args.s3_base_path,
            s3_bucket=args.s3_bucket_name,
            hosts=env.hosts,
            keyspaces=env.keyspaces,
            table=args.table
        )
        worker.snapshot(snapshot)
    else:
        logging.info("Add incrementals to snapshot {!s}".format(
            existing_snapshot))
        worker.update_snapshot(existing_snapshot)


def list_backups(args):
    snapshots = SnapshotCollection(
        args.aws_access_key_id,
        args.aws_secret_access_key,
        args.s3_base_path,
        args.s3_bucket_name,
        get_s3_connection_host(args.s3_bucket_region)
    )
    path_snapshots = defaultdict(list)

    for snapshot in snapshots:
        base_path = os.path.join(snapshot.base_path.split('/')[:-1])
        path_snapshots[base_path].append(snapshot)

    for path, snapshots in path_snapshots.iteritems():
        for snapshot in snapshots:
            print("\t {!r} hosts:{!r} keyspaces:{!r} table:{!r}".format(
                snapshot, snapshot.hosts, snapshot.keyspaces, snapshot.table))
        print("------------------------{}".format('-' * len(path)))


def restore_backup(args):

    _set_user_env_vars(args)
    env.host_string = args.host

    snapshots = SnapshotCollection(
        args.aws_access_key_id,
        args.aws_secret_access_key,
        args.s3_base_path,
        args.s3_bucket_name,
        get_s3_connection_host(args.s3_bucket_region)
    )

    if args.snapshot_name == 'LATEST':
        snapshot = snapshots.get_latest()
    else:
        snapshot = snapshots.get_snapshot_by_name(args.snapshot_name)

    worker = RestoreWorker(aws_access_key_id=args.aws_access_key_id,
                           aws_secret_access_key=args.aws_secret_access_key,
                           s3_bucket_region=args.s3_bucket_region,
                           snapshot=snapshot,
                           cassandra_tools_bin_dir=args.cassandra_tools_bin_dir,
                           restore_dir=args.restore_dir,
                           use_sudo=args.use_sudo)

    worker.restore(args.keyspace)


def main():
    a_base_parser = add_s3_arguments(base_parser)
    subparsers = a_base_parser.add_subparsers(
        title='subcommands', dest='subcommand'
    )

    subparsers.add_parser('list', help="List existing backups")

    backup_parser = subparsers.add_parser('backup', help="Create a snapshot")

    # snapshot / backup arguments
    backup_parser.add_argument(
        '--buffer-size',
        default=64,
        help="The buffer size (MB) for compress and upload")

    backup_parser.add_argument(
        '--exclude-tables',
        default='',
        help="Column families you want to skip")

    backup_parser.add_argument(
        '--hosts',
        required=True,
        help="Comma separated list of hosts to snapshot")

    backup_parser.add_argument(
        '--keyspaces',
        default='',
        help="Comma separated list of keyspaces to backup (omit to backup all)")

    backup_parser.add_argument(
        '--table',
        default='',
        help="The table (column family) to backup")

    backup_parser.add_argument(
        '--cassandra-conf-path',
        default='/etc/cassandra/',
        help="cassandra config file path")

    backup_parser.add_argument(
        '--cassandra-tools-bin-dir',
        default='/usr/bin',
        help="binary directory for nodetool, cqlsh and sstableloader")

    backup_parser.add_argument(
        '--user',
        help="The ssh user to logging on nodes")

    backup_parser.add_argument(
        '--use-sudo',
        default=False,
        help="Use sudo to run backup")

    backup_parser.add_argument(
        '--sudo-user',
        default=None,
        help="Run sudo as an user other than the default (root)")

    backup_parser.add_argument(
        '--sshport',
        help="The ssh port to use to connect to the nodes")

    backup_parser.add_argument(
        '--password',
        default='',
        help="User password to connect with hosts")

    backup_parser.add_argument(
        '--sshkey',
        help="The file containing the private ssh key to use to connect with hosts")

    backup_parser.add_argument(
        '--new-snapshot',
        action='store_true',
        help="Create a new snapshot")

    backup_parser.add_argument(
        '--backup-schema',
        action='store_true',
        help="Backup (thrift) schema of selected keyspaces")

    backup_parser.add_argument(
        '--cqlsh-user',
        default='',
        help="User to use for cqlsh commands")

    backup_parser.add_argument(
        '--cqlsh-password',
        default='',
        help="Password to use for cqlsh commands")

    backup_parser.add_argument(
        '--connection-pool-size',
        default=12,
        help="Number of simultaneous connections to cassandra nodes")

    backup_parser.add_argument(
        '--reduced-redundancy',
        action='store_true',
        help="Use S3 reduced redundancy")

    backup_parser.add_argument(
        '--rate-limit',
        default=0,
        help="Limit the upload speed to S3 (by using 'pv'). Value expressed in kilobytes (*1024)")

    backup_parser.add_argument(
        '--quiet',
        action='store_true',
        help="Set pv in quiet mode when using --rate-limit. "
             "Useful when called by a script.")

    # restore snapshot arguments
    restore_parser = subparsers.add_parser(
        'restore', help="Restores a snapshot")

    restore_parser.add_argument(
        '--snapshot-name',
        default='LATEST',
        help="The name of the snapshot (and incremental backups) to restore")

    restore_parser.add_argument(
        '--keyspace',
        required=True,
        help="The keyspace to restore")

    restore_parser.add_argument(
        '--user',
        help="The ssh user to logging on nodes")

    restore_parser.add_argument(
        '--use-sudo',
        default=False,
        help="Use sudo to restore the backup")

    restore_parser.add_argument(
        '--sudo-user',
        default=None,
        help="Run sudo as an user other than the default (root)")

    restore_parser.add_argument(
        '--sshport',
        help="The ssh port to use to connect to the nodes")

    restore_parser.add_argument(
        '--password',
        default='',
        help="User password to connect with hosts")

    restore_parser.add_argument(
        '--sshkey',
        help="The file containing the private ssh key to use to connect with hosts")

    restore_parser.add_argument(
        '--host',
        required=True,
        help="Cassandra host to restore into")

    restore_parser.add_argument(
        '--cassandra-tools-bin-dir',
        default='/usr/bin',
        help="binary directory for nodetool, cqlsh and sstableloader")

    restore_parser.add_argument(
        '--restore-dir',
        default='/tmp/cassandra_restore',
        help="directory to recover the backup to")

    args = a_base_parser.parse_args()
    subcommand = args.subcommand

    if args.verbose:
        logging.basicConfig(level=logging.INFO)

    if subcommand == 'backup':
        run_backup(args)
    elif subcommand == 'list':
        list_backups(args)
    elif subcommand == 'restore':
        restore_backup(args)


if __name__ == '__main__':
    main()



from functools import wraps
import errno
import os
import signal


class TimeoutError(Exception):
    pass


def timeout(seconds=10, error_message=os.strerror(errno.ETIME)):
    def decorator(func):
        def _handle_timeout(signum, frame):
            raise TimeoutError(error_message)

        @wraps(func)
        def wrapper(*args, **kwargs):
            signal.signal(signal.SIGALRM, _handle_timeout)
            signal.alarm(seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result

        return wrapper

    return decorator



import datetime
import errno
import logging
import os

from logging import handlers
from os import path


# Global logging handlers created by configure.
HANDLERS = []


class IndentFormatter(logging.Formatter):
    def format(self, record, *args, **kwargs):
        """
        Format a message in the log
        Act like the normal format, but indent anything that is a
        newline within the message.
        """
        return logging.Formatter.format(
            self, record, *args, **kwargs).replace('\n', '\n' + ' ' * 8)


def configure(*args, **kwargs):
    """
    Configure logging.
    Borrowed from logging.basicConfig
    Uses the IndentFormatter instead of the regular Formatter
    Also, opts the caller into Syslog output, unless syslog could not
    be opened for some reason or another, in which case a warning will
    be printed to the other log handlers.
    """
    # Configuration must only happen once: no mechanism for avoiding
    # duplication of handlers exists.
    assert len(HANDLERS) == 0

    log_destinations = get_log_destinations()

    if 'stderr' in log_destinations:
        # Add stderr output.
        HANDLERS.append(logging.StreamHandler())

    def terrible_log_output(s):
        import sys

        print >> sys.stderr, s

    places = [
        # Linux
        '/dev/log',

        # FreeBSD
        '/var/run/log',

        # Macintosh
        '/var/run/syslog',
    ]

    default_syslog_address = places[0]
    for p in places:
        if path.exists(p):
            default_syslog_address = p
            break

    syslog_address = kwargs.setdefault('syslog_address',
                                       default_syslog_address)

    valid_facility = False
    if 'syslog' in log_destinations:
        facility, valid_facility = get_syslog_facility()

        if not valid_facility:
            terrible_log_output('invalid syslog facility level specified')

        try:
            # Add syslog output.
            HANDLERS.append(handlers.SysLogHandler(syslog_address,
                                                   facility=facility))
        except EnvironmentError, e:
            if e.errno in [errno.EACCES, errno.ECONNREFUSED]:
                message = ('cassandra-snapshotter: Could not set up syslog, '
                           'continuing anyway.  '
                           'Reason: {0}').format(errno.errorcode[e.errno])

                terrible_log_output(message)

    fs = kwargs.get("format", logging.BASIC_FORMAT)
    dfs = kwargs.get("datefmt", None)
    fmt = IndentFormatter(fs, dfs)

    for handler in HANDLERS:
        handler.setFormatter(fmt)
        logging.root.addHandler(handler)

    # Default to INFO level logging.
    set_level(kwargs.get('level', logging.INFO))


def get_log_destinations():
    """Parse env string"""
    # if env var is not set default to stderr + syslog
    env = os.getenv('CASSANDRA_SNAPSHOTTER_LOG_DESTINATION', 'stderr,syslog')
    return env.split(",")


def get_syslog_facility():
    """Get syslog facility from ENV var"""
    facil = os.getenv('CASSANDRA_SNAPSHOTTER_SYSLOG_FACILITY', 'user')

    valid_facility = True
    try:
        facility = handlers.SysLogHandler.facility_names[facil.lower()]
    except KeyError:
        valid_facility = False
        facility = handlers.SysLogHandler.LOG_USER

    return facility, valid_facility


def set_level(level):
    """Adjust the logging level of CASSANDRA_SNAPSHOTTER"""
    for handler in HANDLERS:
        handler.setLevel(level)

    logging.root.setLevel(level)


class CassandraSnapshotterLogger(object):
    def __init__(self, *args, **kwargs):
        self._logger = logging.getLogger(*args, **kwargs)

    @staticmethod
    def _fmt_structured(d):
        """Formats '{k1:v1, k2:v2}' => 'time=... pid=... k1=v1 k2=v2'
        Output is lexically sorted, *except* the time and pid always
        come first, to assist with human scanning of the data.
        """
        timeEntry = datetime.datetime.utcnow().strftime(
            "time=%Y-%m-%dT%H:%M:%S.%f-00")
        pidEntry = "pid=" + str(os.getpid())

        rest = sorted('='.join([unicode(k), unicode(v)])
                      for (k, v) in d.items())

        return ' '.join([timeEntry, pidEntry] + rest)

    @staticmethod
    def fmt_logline(msg, detail=None, hint=None, structured=None):
        msg_parts = ['MSG: ' + msg]

        if detail is not None:
            msg_parts.append('DETAIL: ' + detail)
        if hint is not None:
            msg_parts.append('HINT: ' + hint)

        if structured is None:
            structured = {}

        msg_parts.append('STRUCTURED: ' +
                         CassandraSnapshotterLogger._fmt_structured(structured))

        return '\n'.join(msg_parts)

    def log(self, level, msg, *args, **kwargs):
        detail = kwargs.pop('detail', None)
        hint = kwargs.pop('hint', None)
        structured = kwargs.pop('structured', None)

        self._logger.log(
            level,
            self.fmt_logline(msg, detail, hint, structured),
            *args, **kwargs)

    def debug(self, *args, **kwargs):
        self.log(logging.DEBUG, *args, **kwargs)

    def info(self, *args, **kwargs):
        self.log(logging.INFO, *args, **kwargs)

    def warning(self, *args, **kwargs):
        self.log(logging.WARNING, *args, **kwargs)

    def error(self, *args, **kwargs):
        self.log(logging.ERROR, *args, **kwargs)

    def critical(self, *args, **kwargs):
        self.log(logging.CRITICAL, *args, **kwargs)



from __future__ import (absolute_import, print_function)

import re
import shutil

import boto
import sys
from boto.s3.connection import S3Connection
from fabric.state import env
from yaml import load

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader
import os
import time
import glob
import logging
import multiprocessing
from multiprocessing.dummy import Pool

from cassandra_backups import logging_helper
from cassandra_backups.utils import (add_s3_arguments, base_parser,
                                     map_wrap, get_s3_connection_host,
                                     check_lzop, check_pv, compressed_pipe,
                                     decompression_pipe)

DEFAULT_CONCURRENCY = max(multiprocessing.cpu_count() - 1, 1)
BUFFER_SIZE = 64  # Default bufsize is 64M
MBFACTOR = float(1 << 20)
MAX_RETRY_COUNT = 4
SLEEP_TIME = 2
SLEEP_MULTIPLIER = 3
UPLOAD_TIMEOUT = 600
DEFAULT_REDUCED_REDUNDANCY = False

logging_helper.configure(
    format='%(name)-12s %(levelname)-8s %(message)s')

logger = logging_helper.CassandraSnapshotterLogger('cassandra_backups.agent')
boto.set_stream_logger('boto', logging.WARNING)


def put_from_manifest(
        s3_bucket, s3_connection_host, s3_ssenc, s3_base_path,
        aws_access_key_id, aws_secret_access_key, manifest,
        bufsize, reduced_redundancy, rate_limit, quiet,
        concurrency=None, incremental_backups=False):
    """
    Uploads files listed in a manifest to amazon S3
    to support larger than 5GB files multipart upload is used (chunks of 60MB)
    files are uploaded compressed with lzop, the .lzo suffix is appended
    """
    exit_code = 0
    bucket = _get_bucket(
        s3_bucket, aws_access_key_id,
        aws_secret_access_key, s3_connection_host)
    manifest_fp = open(manifest, 'r')
    buffer_size = int(bufsize * MBFACTOR)
    files = manifest_fp.read().splitlines()
    pool = Pool(concurrency)
    for f in pool.imap(_upload_file,
                       ((bucket, f, _destination_path(s3_base_path, f), s3_ssenc, buffer_size, reduced_redundancy,
                         rate_limit, quiet)
                        for f in files if f)):
        if f is None:
            # Upload failed.
            exit_code = 1
        elif incremental_backups:
            # Delete files that were successfully uploaded.
            os.remove(f)
    pool.terminate()
    exit(exit_code)


@map_wrap
def _upload_file(bucket, source, destination, s3_ssenc, bufsize, reduced_redundancy, rate_limit, quiet):
    mp = None
    retry_count = 0
    sleep_time = SLEEP_TIME
    while True:
        try:
            if mp is None:
                # Initiate the multi-part upload.
                try:
                    mp = bucket.initiate_multipart_upload(destination, encrypt_key=s3_ssenc,
                                                          reduced_redundancy=reduced_redundancy)
                    logger.info("Initialized multipart upload for file {!s} to {!s}".format(source, destination))
                except Exception as exc:
                    logger.error(
                        "Error while initializing multipart upload for file {!s} to {!s}".format(source, destination))
                    logger.error(exc.message)
                    raise

            try:
                for i, chunk in enumerate(compressed_pipe(source, bufsize, rate_limit, quiet)):
                    mp.upload_part_from_file(chunk, i + 1, cb=s3_progress_update_callback)
            except Exception as exc:
                logger.error("Error uploading file {!s} to {!s}".format(source, destination))
                logger.error(exc.message)
                raise

            try:
                mp.complete_upload()
            except Exception as exc:
                logger.error("Error completing multipart file upload for file {!s} to {!s}".format(source, destination))
                logger.error(exc.message)
                # The multi-part object may be in a bad state.  Extract an error
                # message if we can, then discard it.
                try:
                    logger.error(mp.to_xml())
                except Exception as exc:
                    pass
                _cancel_upload(bucket, mp, destination)
                mp = None
                raise

            # Successful upload, return the uploaded file.
            return source
        except Exception as exc:
            # Failure anywhere reaches here.
            retry_count += 1
            if retry_count > MAX_RETRY_COUNT:
                logger.error("Retried too many times uploading file {!s}".format(source))
                # Abort the multi-part upload if it was ever initiated.
                if mp is not None:
                    _cancel_upload(bucket, mp, destination)
                return None
            else:
                logger.info("Sleeping before retry")
                time.sleep(sleep_time)
                sleep_time *= SLEEP_MULTIPLIER
                logger.info("Retrying {}/{}".format(retry_count, MAX_RETRY_COUNT))
                # Go round again.


def _get_bucket(
        s3_bucket, aws_access_key_id,
        aws_secret_access_key, s3_connection_host):
    connection = S3Connection(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        host=s3_connection_host
    )
    return connection.get_bucket(s3_bucket, validate=False)


def _destination_path(s3_base_path, file_path, compressed=True):
    suffix = compressed and '.lzo' or ''
    return '/'.join([s3_base_path, file_path + suffix])


def _cancel_upload(bucket, mp, remote_path):
    """
    Safe way to cancel a multipart upload
    sleeps SLEEP_TIME seconds and then makes sure that there are not parts left
    in storage
    """
    attempts = 0
    while attempts < 5:
        try:
            time.sleep(SLEEP_TIME)
            mp.cancel_upload()
            time.sleep(SLEEP_TIME)
            for mp in bucket.list_multipart_uploads():
                if mp.key_name == remote_path:
                    mp.cancel_upload()
            return
        except Exception:
            logger.error("Error while cancelling multipart upload")
            attempts += 1


def create_upload_manifest(
        snapshot_name, snapshot_keyspaces, snapshot_table,
        conf_path, manifest_path, exclude_tables, incremental_backups=False):
    if snapshot_keyspaces:
        keyspace_globs = snapshot_keyspaces.split(',')
    else:
        keyspace_globs = ['*']

    if snapshot_table:
        table_glob = snapshot_table
    else:
        table_glob = '*'

    data_paths = _get_data_path(conf_path)
    files = []
    exclude_tables_list = exclude_tables.split(',')
    for data_path in data_paths:
        for keyspace_glob in keyspace_globs:
            path = [
                data_path,
                keyspace_glob,
                table_glob
            ]
            if incremental_backups:
                path += ['backups']
            else:
                path += ['snapshots', snapshot_name]
            path += ['*']

            path = os.path.join(*path)
            if len(exclude_tables_list) > 0:
                for f in glob.glob(os.path.join(path)):
                    # The current format of a file path looks like:
                    # /var/lib/cassandra/data03/system/compaction_history/snapshots/20151102182658/system-compaction_history-jb-6684-Summary.db
                    if f.split('/')[-4] not in exclude_tables_list:
                        files.append(f.strip())
            else:
                files.append(f.strip() for f in glob.glob(os.path.join(path)))

    with open(manifest_path, 'w') as manifest:
        manifest.write('\n'.join("%s" % f for f in files))


def _get_data_path(conf_path):
    """Retrieve cassandra data_file_directories from cassandra.yaml"""
    config_file_path = os.path.join(conf_path, 'cassandra.yaml')
    with open(config_file_path, 'r') as f:
        cassandra_configs = load(f, Loader=Loader)
    data_paths = cassandra_configs['data_file_directories']
    return data_paths


def local_restore(keyspace, snapshot_path, aws_access_key_id, aws_secret_access_key, host,
                  s3_host, s3_bucket_name, cassandra_tools_bin_dir, restore_dir):
    matcher_string = "(%(host)s).*/(%(keyspace)s)/(%(table)s-[A-Za-z0-9]*)/" % dict(
        host=host, keyspace=keyspace, table=".*?")
    keyspace_path = "/".join([restore_dir, keyspace])
    keyspace_table_matcher = re.compile(matcher_string)

    s3connection = S3Connection(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        host=s3_host)

    keys = []
    tables = set()

    bucket = s3connection.get_bucket(s3_bucket_name, validate=False)

    for k in bucket.list(snapshot_path):
        logging.info(k)
        logging.info(k.name)
        r = keyspace_table_matcher.search(k.name)
        if not r:
            continue

        tables.add(r.group(3))
        keys.append(k)

    _delete_old_dir_and_create_new(keyspace_path, tables)
    total_size = reduce(lambda s, k: s + k.size, keys, 0)

    logging.info("Found %(files_count)d files, with total size \
        of %(size)s." % dict(files_count=len(keys),
                             size=_human_size(total_size)))
    print("Found %(files_count)d files, with total size \
        of %(size)s." % dict(files_count=len(keys),
                             size=_human_size(total_size)))

    _download_keys(keys, total_size, restore_dir, keyspace_table_matcher)

    logging.info("Finished downloading...")

    _run_sstableloader(keyspace_path, tables, host, cassandra_tools_bin_dir)


def _delete_old_dir_and_create_new(keyspace_path, tables):
    if os.path.exists(keyspace_path) and os.path.isdir(keyspace_path):
        logging.warning("Deleting directory ({!s})...".format(
            keyspace_path))
        shutil.rmtree(keyspace_path)

    for table in tables:
        path = "{!s}/{!s}".format(keyspace_path, table)
        if not os.path.exists(path):
            os.makedirs(path, mode=0777)


def _download_keys(keys, total_size, restore_dir, keyspace_table_matcher, pool_size=5):
    logging.info("Starting to download...")

    progress_string = ""
    read_bytes = 0

    thread_pool = Pool(pool_size)

    download_key = lambda key: _download_key(restore_dir, keyspace_table_matcher, key)
    for size in thread_pool.imap(download_key, keys):
        old_width = len(progress_string)
        read_bytes += size
        progress_string = "{!s} / {!s} ({:.2f})".format(
            _human_size(read_bytes),
            _human_size(total_size),
            (read_bytes / float(total_size)) * 100.0)
        width = len(progress_string)
        padding = ""
        if width < old_width:
            padding = " " * (width - old_width)
        progress_string = "{!s}{!s}\r".format(progress_string, padding)

        sys.stderr.write(progress_string)


def _download_key(restore_dir, keyspace_table_matcher, key):
    r = keyspace_table_matcher.search(key.name)
    keyspace = r.group(2)
    table = r.group(3)
    host = key.name.split('/')[2]
    if env.host_string != host:
        logging.info("saved host %s doesn't match current host %s , skipping" % (host, env.host_string))  # TODO check
    file_ = key.name.split('/')[-1]
    filename = "{}/{}/{}".format(keyspace, table, file_)
    key_full_path = os.path.join(restore_dir, filename)

    if filename.endswith('.lzo'):
        uncompressed_key_full_path = re.sub('\.lzo$', '', key_full_path)
        logging.info("Decompressing %s..." % key_full_path)
        lzop_pipe = decompression_pipe(uncompressed_key_full_path)
        try:
            key.open_read()
            for chunk in key:
                lzop_pipe.stdin.write(chunk)
            key.close()
            out, err = lzop_pipe.communicate()
            errcode = lzop_pipe.returncode
            if errcode != 0:
                logging.exception("lzop Out: %s\nError:%s\nExit Code %d: " % (out, err, errcode))
        except Exception as e:
            logging.error('LZOP crashed decompressing "{!s}": {!s} - Could be that the file exists already'.format(
                key_full_path, e))
    else:
        try:
            logging.info("Saving %s..." % key_full_path)
            key.get_contents_to_filename(key_full_path)
        except Exception as e:
            logging.error('Unable to create "{!s}": {!s}'.format(key_full_path, e))

    return key.size


def _human_size(size):
    for x in ['bytes', 'KB', 'MB', 'GB']:
        if size < 1024.0:
            return "{:3.1f}{!s}".format(size, x)
        size /= 1024.0
    return "{:3.1f}{!s}".format(size, 'TB')


def _run_sstableloader(keyspace_path, tables, host, cassandra_tools_bin_dir):
    sstableloader = "{!s}/sstableloader".format(cassandra_tools_bin_dir)
    for table in tables:
        path = "/".join([keyspace_path, table])
        if not os.path.exists(path):
            os.makedirs(path, mode=0777)
        command = '%(sstableloader)s --nodes %(hosts)s -v \
            %(keyspace_path)s/%(table)s' % dict(sstableloader=sstableloader, hosts=host,  # FIXME ','.join(hosts),
                                                keyspace_path=keyspace_path, table=table)
        logging.info("invoking: {!s}".format(command))
        os.system(command)


def s3_progress_update_callback(*args):
    # required
    pass


def main():
    subparsers = base_parser.add_subparsers(
        title='subcommands', dest='subcommand')
    base_parser.add_argument(
        '--incremental_backups', action='store_true', default=False)

    put_parser = subparsers.add_parser(
        'put', help="put files on s3 from a manifest")
    manifest_parser = subparsers.add_parser(
        'create-upload-manifest', help="put files on s3 from a manifest")
    fetch_parser = subparsers.add_parser(
        'fetch', help="fetch files from s3")

    # put arguments
    put_parser = add_s3_arguments(put_parser)
    put_parser.add_argument('--bufsize', required=False, type=int, default=BUFFER_SIZE,
                            help="Compress and upload buffer size")
    put_parser.add_argument('--manifest', required=True,
                            help="The manifest containing the files to put on s3")
    put_parser.add_argument('--concurrency', required=False, type=int, default=DEFAULT_CONCURRENCY,
                            help="Compress and upload concurrent processes")
    put_parser.add_argument('--reduced-redundancy', required=False, default=DEFAULT_REDUCED_REDUNDANCY,
                            help="Compress and upload concurrent processes", action="store_true")
    put_parser.add_argument('--rate-limit', required=False, type=int, default=0,
                            help="Limit the upload speed to S3 (by using 'pv'). Value expressed in kilobytes (*1024)")
    put_parser.add_argument('--quiet',
                            help="pv quiet mode, useful when called by a script.", action='store_true')

    # create-upload-manifest arguments
    manifest_parser.add_argument('--snapshot_name', required=True, type=str)
    manifest_parser.add_argument('--conf_path', required=True, type=str)
    manifest_parser.add_argument('--manifest_path', required=True, type=str)
    manifest_parser.add_argument('--snapshot_keyspaces', required=False, type=str, default='')
    manifest_parser.add_argument('--snapshot_table', required=False, type=str, default='')
    manifest_parser.add_argument('--exclude_tables', required=False, type=str)

    # fetch arguments
    fetch_parser.add_argument('--keyspace', required=True, type=str)
    fetch_parser.add_argument('--snapshot-path', required=True, type=str)
    fetch_parser.add_argument('--aws-access-key-id', required=True, type=str)
    fetch_parser.add_argument('--aws-secret-access-key', required=True, type=str)
    fetch_parser.add_argument('--s3-host', required=True, type=str)
    fetch_parser.add_argument('--s3-bucket-name', required=True, type=str)
    fetch_parser.add_argument('--host', required=True, type=str)
    fetch_parser.add_argument('--cassandra-tools-bin-dir', required=True, type=str)
    fetch_parser.add_argument('--restore-dir', required=True, type=str)

    args = base_parser.parse_args()
    subcommand = args.subcommand

    if subcommand == 'create-upload-manifest':
        create_upload_manifest(
            args.snapshot_name,
            args.snapshot_keyspaces,
            args.snapshot_table,
            args.conf_path,
            args.manifest_path,
            args.exclude_tables,
            args.incremental_backups
        )

    if subcommand == 'put':
        check_lzop()

        if args.rate_limit > 0:
            check_pv()

        put_from_manifest(
            args.s3_bucket_name,
            get_s3_connection_host(args.s3_bucket_region),
            args.s3_ssenc,
            args.s3_base_path,
            args.aws_access_key_id,
            args.aws_secret_access_key,
            args.manifest,
            args.bufsize,
            args.reduced_redundancy,
            args.rate_limit,
            args.quiet,
            args.concurrency,
            args.incremental_backups
        )

    if subcommand == 'fetch':
        check_lzop()
        local_restore(args.keyspace,
                      args.snapshot_path,
                      args.aws_access_key_id,
                      args.aws_secret_access_key,
                      args.host,
                      args.s3_host,
                      args.s3_bucket_name,
                      args.cassandra_tools_bin_dir,
                      args.restore_dir)

if __name__ == '__main__':
    check_lzop()



from __future__ import (absolute_import, print_function)

import time
import json
import logging
from distutils.util import strtobool
from boto.s3.connection import S3Connection
from boto.s3.key import Key
from boto.exception import S3ResponseError
from datetime import datetime
from fabric.api import (env, execute, hide, run, sudo)
from fabric.context_managers import settings

from cassandra_backups.utils import get_s3_connection_host


class Snapshot(object):
    """
    A Snapshot instance keeps the details about a cassandra snapshot

    Multiple snapshots can be stored in a single S3 bucket

    A Snapshot is best described by:
        - its name (which defaults to the utc time of creation)
        - the list of hostnames the snapshot runs on
        - the list of keyspaces being backed up
        - the keyspace table being backed up
        - the S3 bucket's base path where the snapshot is stored

    Snapshots data (and incremental backups) are stored using the
    following convention:

        s3_bucket_name:/<base_path>/<snapshot_name>/<node-hostname>/...

    Snapshots are represented on S3 by their manifest file, this makes
    incremental backups much easier
    """

    SNAPSHOT_TIMESTAMP_FORMAT = '%Y%W'

    def __init__(self, base_path, s3_bucket, hosts, keyspaces, table):
        self.s3_bucket = s3_bucket
        self.name = self.make_snapshot_name()
        self.hosts = hosts
        self.keyspaces = keyspaces
        self.table = table
        self._base_path = base_path

    def dump_manifest_file(self):
        manifest_data = {
            'name': self.name,
            'base_path': self._base_path,
            'hosts': self.hosts,
            'keyspaces': self.keyspaces,
            'table': self.table
        }
        return json.dumps(manifest_data)

    @staticmethod
    def load_manifest_file(data, s3_bucket):
        manifest_data = json.loads(data)
        snapshot = Snapshot(
            base_path=manifest_data['base_path'],
            s3_bucket=s3_bucket,
            hosts=manifest_data['hosts'],
            keyspaces=manifest_data['keyspaces'],
            table=manifest_data['table']
        )
        snapshot.name = manifest_data['name']
        return snapshot

    @property
    def base_path(self):
        return '/'.join([self._base_path, self.name])

    @staticmethod
    def make_snapshot_name():
        return datetime.utcnow().strftime(Snapshot.SNAPSHOT_TIMESTAMP_FORMAT)

    def unix_time_name(self):
        dt = datetime.strptime(self.name, self.SNAPSHOT_TIMESTAMP_FORMAT)
        return time.mktime(dt.timetuple()) * 1000

    def __cmp__(self, other):
        return self.unix_time_name() - other.unix_time_name()

    def __repr__(self):
        return self.name

    __str__ = __repr__


class RestoreWorker(object):
    def __init__(self, aws_access_key_id, aws_secret_access_key, s3_bucket_region, snapshot,
                 cassandra_tools_bin_dir, restore_dir, use_sudo):
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_access_key_id = aws_access_key_id
        self.s3_host = get_s3_connection_host(s3_bucket_region)
        self.snapshot = snapshot
        self.cassandra_tools_bin_dir = cassandra_tools_bin_dir
        self.restore_dir = restore_dir
        self.use_sudo = use_sudo

    def restore(self, keyspace):

        logging.info("Restoring keyspace=%(keyspace)s to host %(host)s ,\
            " % dict(keyspace=keyspace, host=env.host_string))

        restore_command = "cassandra-backups-agent " \
                          "fetch " \
                          "--keyspace=%(keyspace)s " \
                          "--snapshot-path=%(snapshot_path)s " \
                          "--aws-access-key-id=%(aws_access_key_id)s " \
                          "--aws-secret-access-key=%(aws_secret_access_key)s  " \
                          "--s3-host=%(s3_host)s  " \
                          "--s3-bucket-name=%(s3_bucket_name)s " \
                          "--host=%(host)s " \
                          "--cassandra-tools-bin-dir=%(cassandra_tools_bin_dir)s " \
                          "--restore-dir=%(restore_dir)s "

        cmd = restore_command % dict(
            keyspace=keyspace,
            snapshot_path=self.snapshot.base_path,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            s3_host=self.s3_host,
            s3_bucket_name=self.snapshot.s3_bucket,
            host=env.host_string,
            cassandra_tools_bin_dir=self.cassandra_tools_bin_dir,
            restore_dir=self.restore_dir,
        )

        if self.use_sudo:
            sudo(cmd)
        else:
            run(cmd)


class BackupWorker(object):
    """
    Backup process is split in this steps:
        - requests cassandra to create new backups
        - uploads backup files to S3
        - clears backup files from nodes
        - updates backup meta information

    When performing a new snapshot the manifest of the snapshot is
    uploaded to S3 for later use.

    Snapshot's manifest path:
    /<snapshot_base_path>/<snapshot_name>/manifest.json

    Every time a backup is done a description of the current ring is
    saved next to the snapshot manifest file

    """

    def __init__(self, aws_secret_access_key,
                 aws_access_key_id, s3_bucket_region, s3_ssenc,
                 s3_connection_host, cassandra_conf_path, use_sudo,
                 cassandra_tools_bin_dir, cqlsh_user, cqlsh_password,
                 backup_schema, buffer_size, exclude_tables, rate_limit, quiet,
                 connection_pool_size=12, reduced_redundancy=False):
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_access_key_id = aws_access_key_id
        self.s3_bucket_region = s3_bucket_region
        self.s3_ssenc = s3_ssenc
        self.s3_connection_host = s3_connection_host
        self.cassandra_conf_path = cassandra_conf_path
        self.nodetool_path = "{!s}/nodetool".format(cassandra_tools_bin_dir)
        self.cqlsh_path = "{!s}/cqlsh".format(cassandra_tools_bin_dir)
        self.cqlsh_user = cqlsh_user
        self.cqlsh_password = cqlsh_password
        self.backup_schema = backup_schema
        self.connection_pool_size = connection_pool_size
        self.buffer_size = buffer_size
        self.reduced_redundancy = reduced_redundancy
        self.rate_limit = rate_limit
        self.quiet = quiet
        if isinstance(use_sudo, basestring):
            self.use_sudo = bool(strtobool(use_sudo))
        else:
            self.use_sudo = use_sudo
        self.exclude_tables = exclude_tables

    def get_current_node_hostname(self):
        return env.host_string

    def upload_node_backups(self, snapshot, incremental_backups):
        prefix = '/'.join(snapshot.base_path.split('/') + [self.get_current_node_hostname()])

        manifest_path = '/tmp/backupmanifest'
        manifest_command = "cassandra-backups-agent " \
                           "%(incremental_backups)s create-upload-manifest " \
                           "--manifest_path=%(manifest_path)s " \
                           "--snapshot_name=%(snapshot_name)s " \
                           "--snapshot_keyspaces=%(snapshot_keyspaces)s " \
                           "--snapshot_table=%(snapshot_table)s " \
                           "--conf_path=%(conf_path)s " \
                           "--exclude_tables=%(exclude_tables)s"
        cmd = manifest_command % dict(
            manifest_path=manifest_path,
            snapshot_name=snapshot.name,
            snapshot_keyspaces=','.join(snapshot.keyspaces or ''),
            snapshot_table=snapshot.table,
            conf_path=self.cassandra_conf_path,
            exclude_tables=self.exclude_tables,
            incremental_backups=incremental_backups and '--incremental_backups' or ''
        )

        if self.use_sudo:
            sudo(cmd)
        else:
            run(cmd)

        upload_command = "cassandra-backups-agent %(incremental_backups)s " \
                         "put " \
                         "--s3-bucket-name=%(bucket)s " \
                         "--s3-bucket-region=%(s3_bucket_region)s %(s3_ssenc)s " \
                         "--s3-base-path=%(prefix)s " \
                         "--manifest=%(manifest)s " \
                         "--bufsize=%(bufsize)s " \
                         "--concurrency=4"

        if self.reduced_redundancy:
            upload_command += " --reduced-redundancy"

        if self.rate_limit > 0:
            upload_command += " --rate-limit=%(rate_limit)s"

        if self.quiet:
            upload_command += " --quiet"

        if self.aws_access_key_id and self.aws_secret_access_key:
            upload_command += " --aws-access-key-id=%(key)s " \
                              "--aws-secret-access-key=%(secret)s"

        cmd = upload_command % dict(
            bucket=snapshot.s3_bucket,
            s3_bucket_region=self.s3_bucket_region,
            s3_ssenc=self.s3_ssenc and '--s3-ssenc' or '',
            prefix=prefix,
            key=self.aws_access_key_id,
            secret=self.aws_secret_access_key,
            manifest=manifest_path,
            bufsize=self.buffer_size,
            rate_limit=self.rate_limit,
            incremental_backups=incremental_backups and '--incremental_backups' or ''
        )

        if self.use_sudo:
            sudo(cmd)
        else:
            run(cmd)

    def snapshot(self, snapshot):
        """
        Perform a snapshot
        """
        logging.info("Create {!r} snapshot".format(snapshot))
        try:
            self.start_cluster_backup(snapshot, incremental_backups=False)
            self.upload_cluster_backups(snapshot, incremental_backups=False)
        finally:
            self.clear_cluster_snapshot(snapshot)
        self.write_ring_description(snapshot)
        self.write_snapshot_manifest(snapshot)
        if self.backup_schema:
            self.write_schema(snapshot)

    def update_snapshot(self, snapshot):
        """Updates backup data changed since :snapshot was done"""
        logging.info("Update {!r} snapshot".format(snapshot))
        self.start_cluster_backup(snapshot, incremental_backups=True)
        self.upload_cluster_backups(snapshot, incremental_backups=True)
        self.write_ring_description(snapshot)
        if self.backup_schema:
            self.write_schema(snapshot)

    def get_ring_description(self):
        with settings(host_string=env.hosts[0]):
            with hide('output'):
                if self.use_sudo:
                    ring_description = sudo(self.nodetool_path + ' ring')
                else:
                    ring_description = run(self.nodetool_path + ' ring')
        return ring_description

    def get_keyspace_schema(self, keyspace=None):
        if self.cqlsh_user and self.cqlsh_password:
            auth = "-u {!s} -p {!s}".format(self.cqlsh_user, self.cqlsh_password)
        else:
            auth = ""
        with settings(host_string=env.hosts[0]):
            with hide('output'):
                cmd = "{!s} {!s} -e 'DESCRIBE SCHEMA;'".format(
                    self.cqlsh_path, auth)
                if keyspace:
                    cmd = "{!s} -k {!s} {!s} -e 'DESCRIBE KEYSPACE {!s};'".format(
                        self.cqlsh_path, keyspace, auth, keyspace)
                if self.use_sudo:
                    output = sudo(cmd)
                else:
                    output = run(cmd)
        return output

    def write_on_S3(self, bucket_name, path, content):
        conn = S3Connection(
            self.aws_access_key_id,
            self.aws_secret_access_key,
            host=self.s3_connection_host)
        bucket = conn.get_bucket(bucket_name, validate=False)
        key = bucket.new_key(path)
        key.set_contents_from_string(content)

    def write_ring_description(self, snapshot):
        logging.info("Writing ring description")
        content = self.get_ring_description()
        ring_path = '/'.join([snapshot.base_path, 'ring'])
        self.write_on_S3(snapshot.s3_bucket, ring_path, content)

    def write_schema(self, snapshot):
        if snapshot.keyspaces:
            for ks in snapshot.keyspaces:
                logging.info("Writing schema for keyspace {!s}".format(ks))
                content = self.get_keyspace_schema(ks)
                schema_path = '/'.join(
                    [snapshot.base_path, "schema_{!s}.cql".format(ks)])
                self.write_on_S3(snapshot.s3_bucket, schema_path, content)
        else:
            logging.info("Writing schema for all keyspaces")
            content = self.get_keyspace_schema()
            schema_path = '/'.join([snapshot.base_path, "schema.cql"])
            self.write_on_S3(snapshot.s3_bucket, schema_path, content)

    def write_snapshot_manifest(self, snapshot):
        content = snapshot.dump_manifest_file()
        manifest_path = '/'.join([snapshot.base_path, 'manifest.json'])
        self.write_on_S3(snapshot.s3_bucket, manifest_path, content)

    def start_cluster_backup(self, snapshot, incremental_backups=False):
        logging.info("Creating snapshots")
        with settings(parallel=True, pool_size=self.connection_pool_size):
            execute(self.node_start_backup, snapshot, incremental_backups)

    def node_start_backup(self, snapshot, incremental_backups):
        """Runs snapshot command on a cassandra node"""

        def run_cmd(cmd):
            with hide('running', 'stdout', 'stderr'):
                if self.use_sudo:
                    sudo(cmd)
                else:
                    run(cmd)

        if incremental_backups:
            backup_command = "%(nodetool)s flush %(keyspace)s %(tables)s"

            if snapshot.keyspaces:
                # flush can only take one keyspace at a time.
                for keyspace in snapshot.keyspaces:
                    cmd = backup_command % dict(
                        nodetool=self.nodetool_path,
                        keyspace=keyspace,
                        tables=snapshot.table or ''
                    )
                    run_cmd(cmd)
            else:
                # If no keyspace then can't provide a table either.
                cmd = backup_command % dict(
                    nodetool=self.nodetool_path,
                    keyspace='',
                    tables=''
                )
                run_cmd(cmd)

        else:
            backup_command = "%(nodetool)s snapshot %(table_param)s \
                -t %(snapshot)s %(keyspaces)s"

            if snapshot.table:
                # Only one keyspace can be specified along with a column family.
                table_param = "-cf {!s}".format(snapshot.table)
                for keyspace in snapshot.keyspaces:
                    cmd = backup_command % dict(
                        nodetool=self.nodetool_path,
                        table_param=table_param,
                        snapshot=snapshot.name,
                        keyspaces=keyspace
                    )
                    run_cmd(cmd)
            else:
                cmd = backup_command % dict(
                    nodetool=self.nodetool_path,
                    table_param='',
                    snapshot=snapshot.name,
                    keyspaces=' '.join(snapshot.keyspaces or '')
                )
                run_cmd(cmd)

    def upload_cluster_backups(self, snapshot, incremental_backups):
        logging.info("Uploading backups")
        with settings(parallel=True, pool_size=self.connection_pool_size):
            execute(self.upload_node_backups, snapshot, incremental_backups)

    def clear_cluster_snapshot(self, snapshot):
        logging.info("Clearing snapshots")
        with settings(parallel=True, pool_size=self.connection_pool_size):
            execute(self.clear_node_snapshot, snapshot)

    def clear_node_snapshot(self, snapshot):
        """Cleans up snapshots from a cassandra node"""
        clear_command = '%(nodetool)s clearsnapshot -t "%(snapshot)s"'
        cmd = clear_command % dict(
            nodetool=self.nodetool_path,
            snapshot=snapshot.name
        )
        if self.use_sudo:
            sudo(cmd)
        else:
            run(cmd)


class SnapshotCollection(object):
    def __init__(
            self, aws_access_key_id,
            aws_secret_access_key, base_path, s3_bucket, s3_connection_host):
        self.s3_connection_host = s3_connection_host
        self.s3_bucket = s3_bucket
        self.base_path = base_path
        self.snapshots = None
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key

    def _read_s3(self):
        if self.snapshots:
            return

        conn = S3Connection(self.aws_access_key_id, self.aws_secret_access_key, host=self.s3_connection_host)
        bucket = conn.get_bucket(self.s3_bucket, validate=False)
        self.snapshots = []
        prefix = self.base_path
        if not self.base_path.endswith('/'):
            prefix = "{!s}/".format(self.base_path)
        snap_paths = [snap.name for snap in bucket.list(
            prefix=prefix, delimiter='/')]
        # Remove the root dir from the list since it won't have a manifest file.
        snap_paths = [x for x in snap_paths if x != prefix]
        for snap_path in snap_paths:
            mkey = Key(bucket)
            manifest_path = '/'.join([snap_path, 'manifest.json'])
            mkey.key = manifest_path
            try:
                manifest_data = mkey.get_contents_as_string()
            except S3ResponseError as e:  # manifest.json not found.
                logging.warn("Response: {!r} manifest_path: {!r}".format(
                    e.message, manifest_path))
                continue
            try:
                self.snapshots.append(
                    Snapshot.load_manifest_file(manifest_data, self.s3_bucket))
            except Exception as e:  # Invalid json format.
                logging.error("Parsing manifest.json failed. {!r}".format(
                    e.message))
                continue
        self.snapshots = sorted(self.snapshots, reverse=True)

    def get_snapshot_by_name(self, name):
        snapshots = filter(lambda s: s.name == name, self)
        return snapshots and snapshots[0]

    def get_latest(self):
        self._read_s3()
        return self.snapshots[0]

    def get_snapshot_for(self, hosts, keyspaces, table, name):
        """Returns the most recent compatible snapshot"""
        for snapshot in self:
            if snapshot.hosts != hosts:
                continue
            if snapshot.keyspaces != keyspaces:
                continue
            if snapshot.table != table:
                continue
            if snapshot.name != name:
                continue
            return snapshot

    def __iter__(self):
        self._read_s3()
        return iter(self.snapshots)



#!/usr/bin/env python

from setuptools import setup, find_packages
from cassandra_backups import __version__


install_requires = [
    'argparse',
    'fabric',
    'boto>=2.29.1',
    'pyyaml'
]

setup(
    name='cassandra_backups',
    version=__version__,
    maintainer='Juan Alvarez',
    maintainer_email='juan@opbeat.com',
    url='http://github.com/jalvz/cassandra_backups',
    description='cassandra_backups is a tool to backup a Cassandra cluster to Amazon S3.',
    packages=find_packages(),
    zip_safe=False,
    install_requires=install_requires,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'cassandra-backups = cassandra_backups.main:main',
            'cassandra-backups-agent = cassandra_backups.agent:main'
        ]
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Topic :: Software Development',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    keywords=['cassandra', 'AWS', 'S3', 'backup']
)



import sys

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
import argparse
import functools
import subprocess

LZOP_BIN = 'lzop'
PV_BIN = 'pv'

S3_CONNECTION_HOSTS = {
    'us-east-1': 's3.amazonaws.com',
    'us-west-2': 's3-us-west-2.amazonaws.com',
    'us-west-1': 's3-us-west-1.amazonaws.com',
    'eu-central-1': 's3-eu-central-1.amazonaws.com',
    'eu-west-1': 's3-eu-west-1.amazonaws.com',
    'ap-southeast-1': 's3-ap-southeast-1.amazonaws.com',
    'ap-southeast-2': 's3-ap-southeast-2.amazonaws.com',
    'ap-northeast-1': 's3-ap-northeast-1.amazonaws.com',
    'sa-east-1': 's3-sa-east-1.amazonaws.com',
    'cn-north-1': 's3.cn-north-1.amazonaws.com.cn'
}

base_parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=__doc__)

base_parser.add_argument('-v', '--verbose',
                         action='store_true',
                         help='increase output verbosity')


def add_s3_arguments(arg_parser):
    """
    Adds common S3 argument to a parser
    """
    arg_parser.add_argument('--aws-access-key-id',
                            help="public AWS access key.")

    arg_parser.add_argument('--aws-secret-access-key',
                            help="S3 secret access key.")

    arg_parser.add_argument('--s3-bucket-region',
                            default='us-east-1',
                            help="S3 bucket region (default us-east-1)")

    arg_parser.add_argument('--s3-ssenc',
                            action='store_true',
                            help="Enable AWS S3 server-side encryption")

    arg_parser.add_argument('--s3-bucket-name',
                            required=True,
                            help="S3 bucket name for backups.")

    arg_parser.add_argument('--s3-base-path',
                            required=True,
                            help="S3 base path for backups.")

    return arg_parser


def get_s3_connection_host(s3_bucket_region):
    return S3_CONNECTION_HOSTS[s3_bucket_region]


def map_wrap(f):
    """
    Fix annoying multiprocessing.imap bug when sending
    *args and **kwargs
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return apply(f, *args, **kwargs)
    return wrapper


def _check_bin(bin_name):
    try:
        subprocess.check_call("{} --version > /dev/null 2>&1".format(bin_name),
                              shell=True)
    except subprocess.CalledProcessError:
        sys.exit("{} not found on path".format(bin_name))


def check_lzop():
    _check_bin(LZOP_BIN)


def check_pv():
    _check_bin(PV_BIN)


def compressed_pipe(path, size, rate_limit, quiet):
    """
    Returns a generator that yields compressed chunks of
    the given file_path

    compression is done with lzop

    """
    lzop = subprocess.Popen(
        (LZOP_BIN, '--stdout', path),
        bufsize=size,
        stdout=subprocess.PIPE
    )

    if rate_limit > 0:
        pv_cmd = [PV_BIN, '--rate-limit', '{}k'.format(rate_limit)]

        if quiet:
            pv_cmd.insert(1, '--quiet')

        pv = subprocess.Popen(
            pv_cmd,
            stdin=lzop.stdout,
            stdout=subprocess.PIPE
        )

    while True:
        if rate_limit > 0:
            chunk = pv.stdout.read(size)
        else:
            chunk = lzop.stdout.read(size)
        if not chunk:
            break
        yield StringIO(chunk)


def decompression_pipe(path):
    lzop = subprocess.Popen(
        (LZOP_BIN, '-d', '-o', path),
        stdin=subprocess.PIPE
    )
    return lzop
